//! `microrm` is a crate providing a lightweight ORM on top of SQLite.
//!
//! Unlike fancier ORM systems, microrm is intended to be extremely lightweight
//! and code-light, which means that by necessity it is opinionated, and thus
//! lacks the power and flexibility of, say, SeaORM or Diesel. In particular,
//! `microrm` currently makes no attempts to provide database migration support.
//!
//! `microrm` provides two components: modeling and querying. The intention is
//! that the modelling is built statically; dynamic models are not directly
//! supported though are possible. However, since by design microrm does not
//! touch database contents for tables not defined in its model, using raw SQL
//! for any needed dynamic components may be a better choice.
//!
//! Querying supports a small subset of SQL expressed as type composition.
//!
//! A simple example using a SQLite table as an (unindexed) key/value store
//! might look something like this:
//! ```
//! use microrm::Entity;
//! #[derive(Debug,Entity,serde::Serialize,serde::Deserialize)]
//! pub struct KVStore {
//!     pub key: String,
//!     pub value: String
//! }
//!
//! let schema = microrm::model::SchemaModel::new().add::<KVStore>();
//! let db = microrm::DB::new_in_memory(schema).unwrap();
//!
//! microrm::query::add(&db, &KVStore {
//!     key: "a_key".to_string(),
//!     value: "a_value".to_string()
//! });
//!
//! let qr = microrm::query::get_one_by(&db, KVStoreColumns::Key, "a_key");
//!
//! assert_eq!(qr.is_some(), true);
//! assert_eq!(qr.as_ref().unwrap().key, "a_key");
//! assert_eq!(qr.as_ref().unwrap().value, "a_value");
//! ```

mod meta;
pub mod model;
pub mod query;

pub use microrm_macros::{Entity, Modelable};

// no need to show the re-exports in the documentation
#[doc(hidden)]
pub mod re_export {
    pub use rusqlite;
    pub use serde;
    pub use serde_json;
}

#[derive(Debug)]
pub enum DBError {
    ConnectFailure,
    NoSchema,
    DifferentSchema,
    DropFailure,
    CreateFailure,
    SanityCheckFailure,
}

impl std::fmt::Display for DBError {
    fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        fmt.write_fmt(format_args!("Database error: {:?}", self))
    }
}

impl std::error::Error for DBError {}

/// SQLite database connection
pub struct DB {
    conn: rusqlite::Connection,
    schema_hash: String,
    schema: model::SchemaModel,
}

impl DB {
    pub fn new(
        schema: model::SchemaModel,
        path: &str,
        allow_recreate: bool,
    ) -> Result<Self, DBError> {
        Self::from_connection(
            rusqlite::Connection::open(path).map_err(|_| DBError::ConnectFailure)?,
            schema,
            allow_recreate,
        )
    }

    /// Mostly for use in tests, but may be useful in some applications as well.
    pub fn new_in_memory(schema: model::SchemaModel) -> Result<Self, DBError> {
        Self::from_connection(
            rusqlite::Connection::open_in_memory().map_err(|_| DBError::ConnectFailure)?,
            schema,
            true,
        )
    }

    pub fn recreate_schema(&self) -> Result<(), DBError> {
        self.create_schema()
    }

    fn from_connection(
        conn: rusqlite::Connection,
        schema: model::SchemaModel,
        allow_recreate: bool,
    ) -> Result<Self, DBError> {
        let sig = Self::calculate_schema_hash(&schema);
        let ret = Self {
            conn,
            schema_hash: sig,
            schema: schema.add::<meta::Metaschema>(),
        };
        ret.check_schema(allow_recreate)?;
        Ok(ret)
    }

    fn calculate_schema_hash(schema: &model::SchemaModel) -> String {
        use sha2::Digest;

        let mut hasher = sha2::Sha256::new();
        schema
            .drop()
            .iter()
            .map(|sql| hasher.update(sql.as_bytes()))
            .count();
        schema
            .create()
            .iter()
            .map(|sql| hasher.update(sql.as_bytes()))
            .count();

        base64::encode(hasher.finalize())
    }

    fn check_schema(&self, allow_recreate: bool) -> Result<(), DBError> {
        let hash = query::get_one_by(self, meta::MetaschemaColumns::Key, "schema_hash");

        if hash.is_none() {
            if !allow_recreate {
                return Err(DBError::NoSchema);
            }
            self.create_schema()?;
        } else if hash.unwrap().value != self.schema_hash {
            if !allow_recreate {
                return Err(DBError::DifferentSchema);
            }
            self.create_schema()?;
        }

        Ok(())
    }

    fn create_schema(&self) -> Result<(), DBError> {
        for ds in self.schema.drop() {
            let prepared = self.conn.prepare(ds);
            prepared
                .unwrap()
                .execute([])
                .map_err(|_| DBError::DropFailure)?;
        }

        for cs in self.schema.create() {
            let prepared = self.conn.prepare(cs);
            prepared
                .unwrap()
                .execute([])
                .map_err(|_| DBError::CreateFailure)?;
        }

        let add_result = query::add(
            self,
            &meta::Metaschema {
                key: "schema_hash".to_string(),
                value: self.schema_hash.clone(),
            },
        );

        assert!(add_result.is_some());

        let sanity_check = query::get_one_by(self, meta::MetaschemaColumns::Key, "schema_hash");
        assert!(sanity_check.is_some());
        assert_eq!(sanity_check.unwrap().value, self.schema_hash);

        Ok(())
    }
}

#[cfg(test)]
mod test {
    use super::DB;

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct S1 {
        an_id: i32,
    }

    fn simple_schema() -> super::model::SchemaModel {
        super::model::SchemaModel::new().add::<S1>()
    }

    #[test]
    fn in_memory_schema() {
        let _db = DB::new_in_memory(simple_schema());
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct S2 {
        #[microrm_foreign]
        parent_id: S1ID,
    }

    #[test]
    fn simple_foreign_key() {
        let db = DB::new_in_memory(super::model::SchemaModel::new().add::<S1>().add::<S2>())
            .expect("Can't connect to in-memory DB");

        let id = crate::query::add(&db, &S1 { an_id: -1 }).expect("Can't add S1");
        let child_id = crate::query::add(&db, &S2 { parent_id: id }).expect("Can't add S2");

        crate::query::get_one_by_id::<S2>(&db, child_id).expect("Can't get S2 instance");
    }
}
